#!/usr/bin/env python

"""
This is the main module of the file integrity checker 'Spot The Difference'.
It has one class (IntegrityChecker) which implements a generic file
integrity checker. Subclasses must be made to make it work with specific
databases.
The 'IntegrityCehcker' class consists of three sections:
- methods that walk around the file system and collect data about files
- database queries
- output formatting methods
"""


import os
import md5, sha
from time import ctime
from getpass import getuser
from sys import exit, argv, exc_info
from traceback import format_exception
from socket import gethostname, gethostbyname


################################################################################
# Constants definitions                                                        #
################################################################################

# File checks may be identified in two ways:
# - a single character, in the configuration file (e.g. '5' for MD5 checksum)
# - a longer string, in table fields (e.g. 'md5', 'sha', 'st_mode')
# The following is a tuple of the available file checks (as single characters)
# in the same order of the table fieflds. It will help interpreting tuples
# returned by 'fetchone()'
checks = tuple("5spidlugzamc")

# The following is a dictionary of the available file checks. Each key is the
# single-character representation of a check; the corresponding value is the
# longer-string representation of the same check. It will help converting from
# one form to the other.
headers = {"5": "md5",        # md5 checksum
           "s": "sha",        # sha1 checksum
           "p": "st_mode",    # permissions
           "i": "st_ino",     # inode number
           "d": "st_dev",     # device number
           "l": "st_nlink",   # number of links
           "u": "st_uid",     # user ID
           "g": "st_gid",     # group ID
           "z": "st_size",    # size
           "a": "st_atime",   # access time
           "m": "st_mtime",   # modification time
           "c": "st_ctime"}   # change time


################################################################################
# Main class                                                                   #
################################################################################

class IntegrityChecker:

    """This class implements a generic file integrity checker

    It is meant to be subclassed to make it interoperate with various
    databases. Most methods should work fine with any database (as long as its
    python module is compliant with the python DB-API). The only exception is
    the open_db() method, which must be overridden by subclasses. This method
    connects to the database and, since each database has specific access
    parameters (e.g. not all databases require a DSN string), it was impossible
    to make a generic, functional, connection method.
    For a complete reference of command line options, check out the Parser
    module.
    Basically, this class walks around the filesystem and performs user-defined
    checks on files and directories.
    If it is meant to 'update' the database, all checks will be recorded to a
    database.
    If it is meant to 'check' the filesystem, all file checks will be compared
    to the ones stored in the database.
    """

    def __init__(self, options, paths, vars_):
        """Initialize command line and configuration parameters"""
        # command line options
        self.options = options

        # Configuration file parameters
        self.paths = paths
        self.vars = vars_

        # Database name (as will be written in the final report). It will be set
        # inside the 'open_db()' method
        self.db_name = None

    def go(self):
        """Open the database connection and the output file and start walking
           around
        """
        if self.options.update == self.options.check:
            # Both update ('-u') and check ('-c') requested (or none)
            exit("Don't know wether to update ('-u') or check ('-c')")

        db, self._table = self.open_db()
        self.cursor = db.cursor()

        try:
            self.outfile = open(self.options.outfile, "w+t")
        except:
            exit("Could not open output file '%s'" % self.options.outfile)

        self.print_headers()
        start_time = ctime()
        self._error_files = []  # List of files that can't be opened or 'stat'ed

        if self.options.check:

            self._missing_files = [] # List of files removed from the filesystem

            self.outfile.write("\nChecking the filesystem:\n\n\n- Modified:\n") 
            # The check() method directly writes in the output file the path
            # of modified files (to save memory). To see if any modified files
            # have been found, the output file's positions before and after
            # the call to check() are compared.
            cur_position = self.outfile.tell()

            self.check()

            if self.outfile.tell() == cur_position:
                self.outfile.write("\n    No files modified\n")

            self.outfile.write("\n\n- Missing:\n\n")
            if not self._missing_files:
                self.outfile.write("    No files missing\n")
            else:
                for file in self._missing_files:
                    self.outfile.write("    %s\n" % file)

            self.outfile.write("\n\n- Added:\n\n")
            cur_position = self.outfile.tell()
            # List of pathnames recorded in the database
            self._known_paths = self.select_path()

            if self.options.verbose:
                print "Looking for new files..."

        else:
            self.outfile.write("\nUpdating Database:\n")
            self.truncate_table()

        self.walk()

        if self.options.check and (self.outfile.tell() == cur_position):
            self.outfile.write("    No new files found\n")
            if self.options.verbose:
                print "No files added"

        if self._error_files:
            self.outfile.write("\n\n  Could not open these files:\n\n")
            for filename, error in self._error_files:
                self.outfile.write("    %s\n      %s\n" % (filename, error))

        end_time = ctime()

        if self.options.update:
            self.outfile.write("\n\nInserted %d files in the database\n" %
                               self.select_count())
            db.commit()

        self.print_footer(start_time, end_time)

        if self.options.email:
            if self.options.verbose:
                print "sending report by email..."
            if self.sendmail():
                if self.options.verbose:
                    print "mail sent"
            else:
                if self.options.verbose:
                    print "mail not sent"

        self.outfile.close()
        db.close()

    def walk(self):
        """Walk around the filesystem"""
        for path, checks in self.paths["root"].items():
            if os.path.isdir(path):
                os.path.walk(path, self.spot, checks)
            else:
                self.spot(checks, path)

    def spot(self, checks, dirname, filenames = []):
        """Determine if a specific path has to be examined

        Check if the directory 'dirname' and its files and subdirectories are
        to be checked according to the configuration file rules.
        Despite of its name, the 'dirname' argument may also be a file.
        Then, if the user is updating the database, each 'approved' path will
        be passed to the update() method.
        If the user is checking the filesystem, instead, this method is only
        used to look for new files, present on the filesystem but not recorded
        in the database.
        """
        if dirname in self.paths["ignore"]:
            # This directory must be ignored
            del filenames[:]
            self.paths["ignore"].remove(dirname)
            return

        elif dirname in self.paths["dollar"].keys():
            # Special checks apply to this directory
            checks = self.paths["dollar"][dirname]
            del self.paths["dollar"][dirname]

        if self.options.check:
            try:
                self._known_paths.remove(dirname)
            except ValueError:
                # This is a new file!
                self.outfile.write("    %s\n" % dirname)
                if not self.options.quiet:
                    print "Added", dirname
        else:
            if self.options.verbose:
                print "Updating", dirname
            self.update(dirname, checks)

        sub_directories = []
        for filename in filenames:

            fullname = os.path.join(dirname, filename)

            if os.path.isdir(fullname):
                # Ignore directories. They will be checked later.
                if dirname in self.paths["equal"]:
                    # All 'dirname' subdirectories must be ignored.
                    # Then, mark this subdirectory for deletion
                    sub_directories.append(filename)
                continue

            elif fullname in self.paths["ignore"]:
                self.paths["ignore"].remove(fullname)
                continue

            if self.options.update and fullname in self.paths["dollar"].keys():
                # this file wants special checks
                if self.options.verbose:
                    print "Updating", filename
                self.update(fullname, self.paths["dollar"][fullname])
                del self.paths["dollar"][fullname]
                continue

            if self.options.check:
                try:
                    self._known_paths.remove(fullname)
                except ValueError:
                    self.outfile.write("    %s\n" % fullname)
                    if not self.options.quiet:
                        print "Added", fullname
            else:
                if self.options.verbose:
                    print "Updating", fullname
                self.update(fullname, checks)

        for d in sub_directories:
            filenames.remove(d)

    def check(self):
        """Compare the current filesystem to the one stored on the database

        Each file in the database is stored together with the output of the
        checks performed against it. For each of these files, the same checks
        are now performed again, and the new output is compared to the old
        'good' one. Only files in the database are checked in this phase. Then,
        only modified and missing files can be detected, not new ones. To detect
        new files, the walk() method is called later on.
        """

        self.select_all()

        while True:
            row = self.cursor.fetchone()

            try:
                filename = row[0]
            except TypeError:
                # End of query result
                break

            try:
                st = os.stat(filename)
            except OSError:
                self._missing_files.append(filename)
                if not self.options.quiet:
                    print "Missing", filename
                continue

            if self.options.verbose:
                print "Checking", filename

            # Filename to write in the report
            fn = "\n    %s\n" % filename
            for check, value in zip(checks, row[1:]):

                if value is None:
                    continue

                elif check not in "5s":
                    new_value = str(getattr(st, headers[check]))

                else:
                    try:
                        f = open(filename, "rb")
                    except IOError:
                        err = format_exception(*exc_info())[-1].split("] ")[1]
                        self._error_files.append((filename, err))
                        if not self.options.quiet:
                            print "Could not open", filename
                        continue

                    if check == "5":
                        cs = md5.new(f.read())
                    else:
                        cs = sha.new(f.read())
                    f.close()
                    new_value = cs.hexdigest()

                if new_value != str(value):
                    # File modified
                    self.write_formatted(fn, check, value, new_value)
                    if not self.options.quiet and fn:
                        # If 'fn' is set, this is the first failed check on
                        # 'filename'
                        print "Modified", filename
                    fn = ""

    def update(self, filename, checks):
        """Perform file checks and insert them in the database"""
        values = {"path": "'%s'" % filename}

        try:
            # though 'filename' certainly exists, os.stat could fail if it is
            # a broken symlink
            st = os.stat(filename)
        except:
            err = format_exception(*exc_info())[-1].split("] ")[1]
            self._error_files.append((filename, err))
            if not self.options.quiet:
                print "Could not stat", filename
            return

        got_error = False  # True if 'filename' can't be opened
        for c in checks:
            if c not in "5s":
                try:
                    value = str(getattr(st, headers[c]))
                except KeyError:
                    exit("'%s': Unknow check '%s'" % (filename, c))
            elif not got_error:
                if os.path.isdir(filename):
                    continue

                try:
                    f = open(filename, "rb")
                except IOError:
                    # Lack of privileges
                    err = format_exception(*exc_info())[-1].split("] ")[1]
                    self._error_files.append((filename, err))
                    if not self.options.quiet:
                        print "Could not open", filename
                    got_error = True
                    continue

                if c == "5":
                    cs = md5.new(f.read())
                else:
                    cs = sha.new(f.read())
                f.close()
                value = "'%s'" % cs.hexdigest()

            values[headers[c]] = value

        self.update_db(values)

                          #### Database queries ####

    def select_all(self):
        """Return the whole content of the table"""
        try:
            self.cursor.execute("SELECT * FROM %s;" % self._table)
        except:
            exit("Could not select from table '%s'" % self._table)

    def select_path(self):
        """Return a list of all the files stored in the table"""
        self.cursor.execute("SELECT path FROM %s;" % self._table)
        return [row[0] for row in self.cursor.fetchall()]

    def select_count(self):
        """Return the number of rows on the table"""
        self.cursor.execute("SELECT COUNT(*) FROM %s;" % self._table)
        return self.cursor.fetchone()[0]

    def truncate_table(self):
        """Truncate the table to get ready for the update"""
        try:
            self.cursor.execute("TRUNCATE TABLE %s;" % self._table)
        except:
            exit("Could not truncate table '%s'" % self._table)

    def update_db(self, values):
        """Insert values into the database. 'values' is a dictionary
        {field_name: value, ...}.
        """
        try:
            self.cursor.execute("INSERT INTO %s (%s) VALUES (%s);" %
                                (self._table,
                                 ", ".join(values.keys()),
                                 ", ".join(values.values())))
        except:
            # Missing permissions or primary key constraint violated
            exit(format_exception(*exc_info())[-1])

                          #### Report formatting ####

    def write_formatted(self, filename, check, old, new):
        """Write modifications occured to a file in a readable form"""
        if check in "amc":
            # Translate the timestamp in something more readable
            old, new = map(ctime, map(float, (old, new)))
        elif check == "p":
            # Convert permissions from string to decimal and
            # from decimal to octal format
            old, new = map(oct, map(int, (old, new)))
        elif check == "z":
            # Specify that the size is in Bytes
            old, new = [str(size) + " Bytes" for size in old, new]

        self.outfile.write("%s      %s: %s -> %s\n" %
                           (filename, check, old, new))

    def print_headers(self):
        """Print the launch parameters at the beginning of the report"""
        header = "%s Spot The Difference %s\n\n" % (("*" * 29), ("*" * 30))
        hostname = gethostname()
        ip_address = gethostbyname(hostname)
        user = getuser()
        cfg_file = os.path.abspath(self.options.cfg_file)
        cmd_line = " ".join(argv)

        self.outfile.write(header                                            +
                           "Date:               %s\n" % ctime()              +
                           "Hostname:           %s\n" % hostname             +
                           "Host address:       %s\n" % ip_address           +
                           "User:               %s\n" % user                 +
                           "Configuration File: %s\n" % cfg_file             +
                           "Database:           %s\n" % self.db_name         +
                           "Database type:      %s\n" % self.vars["db_type"] +
                           "Command line:       %s\n" % cmd_line             +
                           "\n%s\n" % ("*" * 80))

    def print_footer(self, start_time, end_time):
        """Print the start and end time at the end of the report"""
        self.outfile.write("\n%s\n\n" % ("*" * 80)         +
                           "Start Time: %s\n" % start_time +
                           "End Time  : %s\n" % end_time   +
                           "\n%s\n" % ("*" * 80))

    def sendmail(self):
        """Send the final report by email

        A few variables set in the configuration file are used:
        - mail_server:     name (or address) of the SMTP server. A port number
                           may be specified in the form 'server:port'. Default
                           is port 25;
        - mail_recipients: a list of whitespace separated e-mail addresses to
                           send the report to.
        """
        import smtplib

        try:
            mail_server = self.vars["mail_server"]
        except KeyError:
            print "Variable 'mail_server' not set in file '%s'" % \
                  self.options.cfg_file
            return

        # Connect to SMTP server
        try:
            outbox = smtplib.SMTP(mail_server)
        except:
            print "Failed to connect to server '%s'" % mail_server
            return
        else:
            if self.options.verbose:
                print "Connected to server '%s'" % mail_server

        # Get the arguments for outbox.sendmail()
        from_ = "SpotTheDifference@%s" % outbox.local_hostname
        try:
            to = self.vars["mail_recipients"].split()
        except KeyError:
            print "Variable 'mail_recipients' not set in file '%s'" % \
                  self.options.cfg_file
            return

        headers = "From: %s\n" % from_
        headers += "To: %s\n" % "; ".join(to)
        headers += "Subject: Spot The Difference - %s\n\n" % ctime()

        # Send mail
        self.outfile.seek(0)
        try:
            outbox.sendmail(from_, to, headers + self.outfile.read())
        except:
            return

        return True

                         #### Methods to override ####

    def open_db(self):
        """Connect to the database

        Return a database object and the name of the table to work on.
        This method must be overridden by subclasses.
        It must also set the value of the db_name attribute (usually a string in
        the form: '<db>.<table> on <host>').
        """
        pass

