#!/usr/bin/env python

"""
This module is part of the file integrity checker 'Spot The Difference'.
It subcalasses the generic class 'IntegrityChecker' to make it interact with
SQLite databases.
It overrides the 'open_db' and 'truncate_table' methods and extends the
'print_footer' method.
"""


import sqlite
from spot import IntegrityChecker
import md5
from sys import exit


class spot(IntegrityChecker):

    """File integrity checker with SQLite database access"""

    def __init__(self, options, paths, vars_):
        """Call the parent __init__ method"""
        IntegrityChecker.__init__(self, options, paths, vars_)

    def open_db(self):
        """Connect to the database

        Only the 'db' parameter in the configuration file is necessary to
        connect to the database. The database object and the table name are
        returned. The 'db_name' attribute is set to the name of the file
        containing the database.
        """
        try:
            self.db_name = self.vars["db"]
        except KeyError:
            exit("Variable 'db' not set in file '%s'" %
                 self.options.cfg_file)

        try:
            db = sqlite.connect(self.db_name, 0600)
        except:
            exit("Could not connect to file '%s'" % self.db_name)

        try:
            table = self.vars["table"]
        except KeyError:
            exit("Variable 'table' not set in file '%s'" %
                 self.options.cfg_file)

        return (db, table)

    def truncate_table(self):
        """Drop the table and recreate it

        SQLite doesn't support the 'TRUNCATE' statement. The only way to
        truncate a table is to drop it and recreate it.
        """
        try:
            self.cursor.execute("DROP TABLE %s;" % self._table)
        except:
            # Table does not exist yet
            pass

        self.cursor.execute("""CREATE TABLE %s (
                                      path     VARCHAR(150) PRIMARY KEY,
                                      md5      VARCHAR(32),
                                      sha      VARCHAR(40),
                                      st_mode  INTEGER,
                                      st_ino   INTEGER,
                                      st_dev   INTEGER,
                                      st_nlink INTEGER,
                                      st_uid   INTEGER,
                                      st_gid   INTEGER,
                                      st_size  INTEGER,
                                      st_atime INTEGER,
                                      st_mtime INTEGER,
                                      st_ctime INTEGER
                               );""" % self._table)

    def print_footer(self, start_time, end_time):
        """Extend the superclass method adding the fingerprint of the database
        file at the end of the report
        """
        IntegrityChecker.print_footer(self, start_time, end_time)

        f = open(self.db_name, "r")
        m = md5.new(f.read())
        f.close()

        self.outfile.write("\nMD5 fingerprint of the database file '%s' is:\n"
                            % self.db_name)
        self.outfile.write("'%s'\n\n%s\n" % (m.hexdigest(), ("*" * 80)))

